# -*- coding: utf-8 -*-
"""
database handlers module.
"""

import pyrin.database.services as database_services
import pyrin.application.services as application_services

application_services.register_after_request_handler(database_services.finalize_transaction)
application_services.register_teardown_request_handler(database_services.cleanup_session)



# -*- coding: utf-8 -*-
"""
database migration manager module.
"""

from os import path

from sqlalchemy import MetaData

import pyrin.database.services as database_services
import pyrin.configuration.services as config_services
import pyrin.application.services as application_services

from pyrin.core.context import DTO, Manager
from pyrin.database.migration.exceptions import EngineBindNameNotFoundError
from pyrin.database.model.base import CoreEntity


class DatabaseMigrationManager(Manager):
    """
    database migration manager class.
    """

    DEFAULT_ENGINE_NAME = 'default'

    def __init__(self):
        """
        initializes an instance of DatabaseMigrationManager.
        """

        super().__init__()

        # a dictionary containing engine to tables map for all tables.
        # in the form of: {Engine engine: list[Table] tables}
        self._engine_to_table_map = DTO()

        # a dictionary containing bind name to metadata map. note that each metadata
        # only contains tables that should be created on relevant bounded engine.
        # the default engine is added with 'default' key.
        # in the form of: {str bind_name: MataDataAdapter metadata}
        self._bind_name_to_metadata_map = DTO()

    def create_all(self):
        """
        creates all entities on database engine.
        """

        for engine, tables in self._engine_to_table_map.items():
            if tables is not None and len(tables) > 0:
                CoreEntity.metadata.create_all(engine, tables)

    def drop_all(self):
        """
        drops all entities on database engine.
        """

        for engine, tables in self._engine_to_table_map.items():
            if tables is not None and len(tables) > 0:
                CoreEntity.metadata.drop_all(engine, tables)

    def _map_engine_to_table(self):
        """
        maps all engines to relevant tables.
        """

        all_tables = DTO(**CoreEntity.metadata.tables)
        for entity, engine in database_services.get_entity_to_engine_map().items():
            if engine not in self._engine_to_table_map:
                self._engine_to_table_map[engine] = []

            self._engine_to_table_map[engine].append(all_tables.pop(entity.table_fullname()))

        # all remaining tables are associated with default engine.
        if len(all_tables) > 0:
            self._engine_to_table_map[database_services.get_default_engine()] = \
                list(all_tables.values())

    def get_connection_urls(self):
        """
        gets all databases connection urls from config store.
        it gets the values from active section of each store.
        in case of sqlite usage, all urls will be made absolute.

        :returns: dict(str bind_name: str connection_url)
        :rtype: dict
        """

        connections = DTO()
        connections[self.DEFAULT_ENGINE_NAME] = config_services.get_active('database',
                                                                           'sqlalchemy_url')
        binds = config_services.get_active_section('database.binds')
        connections.update(**binds)

        for name, url in connections.items():
            connections[name] = self._get_absolute_connection_url(url)

        return connections

    def _map_bind_to_metadata(self):
        """
        maps bind names of different engines to a
        metadata representing the related tables.
        """

        for engine, tables in self._engine_to_table_map.items():
            bounded_engines = database_services.get_bounded_engines()
            bind_name_map = None
            for bind_name, bounded_engine in bounded_engines.items():
                if engine == bounded_engine:
                    bind_name_map = bind_name
                    break

            default_engine = database_services.get_default_engine()
            if bind_name_map is None and engine == default_engine:
                bind_name_map = self.DEFAULT_ENGINE_NAME
            elif bind_name_map is None and engine != default_engine:
                raise EngineBindNameNotFoundError('Could not find any bind name '
                                                  'for database engine [{name}].'
                                                  .format(name=str(engine)))

            self._bind_name_to_metadata_map[bind_name_map] = self._add_to_metadata(tables)

    def get_bind_name_to_metadata_map(self):
        """
        gets bind name to metadata map.

        :returns: dict(str bind_name: MetaData metadata)
        :rtype: dict
        """

        return self._bind_name_to_metadata_map

    def _add_to_metadata(self, tables):
        """
        adds given tables into a metadata and returns it.

        :param list[Tables] tables: tables to be added into metadata.

        :rtype: MetaData
        """

        metadata = MetaData()
        for table in tables:
            table.tometadata(metadata)

        return metadata

    def configure_migration_data(self):
        """
        configures the required data for any migration.
        """

        self._map_engine_to_table()
        self._map_bind_to_metadata()

    def _get_absolute_connection_url(self, connection_url):
        """
        gets absolute path of given connection string if required.
        it is only required when sqlite is in use.

        :param str connection_url: connection url.

        :rtype: str
        """

        if connection_url is None or 'sqlite' not in connection_url.lower() or \
                connection_url.lower().startswith('sqlite:////') or \
                'sqlite:///:memory:' in connection_url.lower():
            return connection_url

        root_path = application_services.get_application_root_path()
        url = connection_url.replace('sqlite:///', '')
        full_path = path.join(root_path, url)
        absolute_path = path.abspath(full_path)

        return 'sqlite:///{absolute_path}'.format(absolute_path=absolute_path)



# -*- coding: utf-8 -*-
"""
database orm query base module.
"""

import inspect

from sqlalchemy import inspection, log
from sqlalchemy.orm import Query
from sqlalchemy.orm.attributes import InstrumentedAttribute

from pyrin.core.globals import LIST_TYPES, _
from pyrin.database.model.base import CoreEntity
from pyrin.database.orm.query.exceptions import ColumnsOutOfScopeError


@inspection._self_inspects
@log.class_logger
class CoreQuery(Query):
    """
    core query class.
    this is the application default query class.
    it extends sqlalchemy `Query` class.
    """

    def __init__(self, entities, session=None, **options):
        """
        initializes an instance of CoreQuery.

        :param tuple entities: entities or columns that are needed for query.
        :param Session session: optional session object to bind this query to it.

        :keyword Union[type, tuple[type]] scope: class type of the entities that this
                                                 query instance will work on. if the
                                                 query is working on multiple entities,
                                                 this value must be a tuple of all class
                                                 types of that entities.

                                                 for example: if you set
                                                 `entities=SomeEntity.id, AnotherEntity.name`
                                                 you should leave `scope=None` to skip validation
                                                 or you could set
                                                 `scope=(SomeEntity, AnotherEntity)`
                                                 this way validation succeeds, but if
                                                 you set `scope=SomeEntity`
                                                 then the query will not be executed
                                                 and an error would be raised.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        scope = options.get('scope', None)
        if scope is not None:
            self._prepare_validation(entities, scope)

        super().__init__(entities, session)

    def _validate_scope(self, entities, scope):
        """
        validates the query for given entities based on provided scope.

        :param tuple entities: entities or columns that needed for query.

        :param Union[type, tuple[type]] scope: class type of the entities that this
                                               query instance will work on. if the
                                               query is working on multiple entities,
                                               this value must be a tuple of all class
                                               types of that entities.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        scope_classes = set(entity for entity in scope if inspect.isclass(entity)
                            and issubclass(entity, CoreEntity))

        requested_classes = set(entity for entity in entities if inspect.isclass(entity)
                                and issubclass(entity, CoreEntity))

        requested_classes_by_column = set(column.class_ for column in entities if
                                          isinstance(column, InstrumentedAttribute)
                                          and issubclass(getattr(column, 'class_', type),
                                                         CoreEntity))

        all_requested_classes = requested_classes.union(requested_classes_by_column)

        if not all_requested_classes.issubset(scope_classes):
            raise ColumnsOutOfScopeError(_('Requested columns are out of scope of this '
                                           'query. Please revise the requested columns.'))

    def _prepare_validation(self, entities, scope):
        """
        prepares entities and scope and validates them if required.

        :param tuple entities: entities or columns that needed for query.

        :param Union[type, tuple[type]] scope: class type of the entities that this
                                               query instance will work on. if the
                                               query is working on multiple entities,
                                               this value must be a tuple of all class
                                               types of that entities.

        :raises ColumnsOutOfScopeError: columns out of scope error.
        """

        if scope is None or entities is None:
            return

        if not isinstance(scope, LIST_TYPES):
            scope = (scope,)

        if not isinstance(entities, LIST_TYPES):
            entities = (entities,)

        if len(scope) <= 0 or len(entities) <= 0:
            return

        self._validate_scope(entities, scope)



# -*- coding: utf-8 -*-
"""
database orm sql component module.
"""

from pyrin.application.decorators import component
from pyrin.application.context import Component
from pyrin.database.orm.sql import DatabaseORMSQLPackage
from pyrin.database.orm.sql.manager import DatabaseORMSQLManager


@component(DatabaseORMSQLPackage.COMPONENT_NAME)
class DatabaseORMSQLComponent(Component, DatabaseORMSQLManager):
    """
    database orm sql component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
utils sqlalchemy module.
"""

from sqlalchemy.orm import lazyload
from sqlalchemy import func
from sqlalchemy.util import lightweight_named_tuple

import pyrin.utils.datetime as datetime_utils

from pyrin.core.context import DTO
from pyrin.core.globals import LIST_TYPES
from pyrin.database.services import get_current_store
from pyrin.utils.exceptions import InvalidRowResultFieldsAndValuesError, \
    FieldsAndValuesCountMismatchError


def entity_to_dict(entity, exposed_only=True):
    """
    converts the given entity into a dict and returns it.
    the result dict only contains the columns of the entity
    which their `hidden` attribute is set to False.

    :param CoreEntity entity: entity to be converted.

    :param bool exposed_only: if set to False, it returns all
                              columns of the entity as dict.
                              if not provided, defaults to True.
    
    :rtype: dict
    """

    if entity is None:
        return DTO()

    return entity.to_dict(exposed_only)


def dict_to_entity(entity_class, **kwargs):
    """
    converts the given keyword arguments into
    an specified entity and returns it.

    :param type entity_class: the result entity class type.

    :rtype: CoreEntity
    """

    result = entity_class()
    result.from_dict(**kwargs)
    return result


def entity_to_dict_list(entities, exposed_only=True):
    """
    converts the given list of entities into a
    list of dicts and returns the result.

    :param list[CoreEntity] entities: list of entities.

    :param bool exposed_only: if set to False, it returns all
                              columns of the entity as dict.
                              if not provided, defaults to True.

    :returns list[dict]
    :rtype list
    """

    results = []
    if entities is None or len(entities) <= 0:
        return results

    for single_entity in entities:
        results.append(entity_to_dict(single_entity, exposed_only))

    return results


def keyed_tuple_to_dict(value):
    """
    converts the given `AbstractKeyedTuple` object into a dict.

    :param AbstractKeyedTuple value: value to be converted.

    :rtype: dict
    """

    if value is None or len(value) <= 0:
        return DTO()

    return DTO(zip(value.keys(), value))


def keyed_tuple_to_dict_list(values):
    """
    converts the given list of `AbstractKeyedTuple` objects into a list of dicts.

    :param list[AbstractKeyedTuple] values: values to be converted.

    :rtype: dict
    """

    results = []
    if values is None or len(values) <= 0:
        return results

    for single_value in values:
        results.append(keyed_tuple_to_dict(single_value))

    return results


def like_both(value):
    """
    gets a copy of string with `%` attached to both
    ends of it to use in like operator.

    :param str value: value to be processed.

    :rtype: str
    """

    if value is None:
        value = ''

    return '%{value}%'.format(value=value)


def like_begin(value):
    """
    gets a copy of string with `%` attached to beginning
    of it to use in like operator.

    :param str value: value to be processed.

    :rtype: str
    """

    if value is None:
        value = ''

    return '%{value}'.format(value=value)


def like_end(value):
    """
    gets a copy of string with `%` attached to end
    of it to use in like operator.

    :param str value: value to be processed.

    :rtype: str
    """

    if value is None:
        value = ''

    return '{value}%'.format(value=value)


def count(query, column=None):
    """
    executes a count query using given query object and returns the result.

    this method generates a sql query like below:
    select count(column)
    from table
    where ...

    you should always use this method instead of sqlalchemy session.query.count()
    because session.query.count() is inefficient.

    :param Query query: sqlalchemy query object.

    :param Column column: column to perform count on it.
                          defaults to `*` if not provided.

    :rtype: int
    """

    func_count = func.count()
    if column is not None:
        func_count = func.count(column)

    statement = query.options(lazyload('*')).statement.with_only_columns(
        [func_count]).order_by(None)

    store = get_current_store()
    return store.execute(statement).scalar()


def add_range_clause(clauses, column, value_lower, value_upper,
                     include_equal_to_lower=True,
                     include_equal_to_upper=True):
    """
    adds range comparison into given clauses using specified inputs.

    :param list clauses: clause list to add range clause to it.
    :param Column column: entity column to add range clause for it.
    :param object value_lower: lower bound of range clause.
    :param object value_upper: upper bound of range clause.

    :param include_equal_to_lower: specifies that lower value
                                   should be considered in range.
                                   defaults to True if not provided.

    :param include_equal_to_upper: specifies that upper value
                                   should be considered in range.
                                   defaults to True if not provided.
    """

    if value_lower is None and value_upper is None:
        return

    # we swap the upper and lower values in case of user mistake.
    if value_lower is not None and value_upper is not None and value_lower > value_upper:
        value_lower, value_upper = value_upper, value_lower

    if value_lower is not None and value_lower == value_upper:
        clauses.append(column == value_lower)
    else:
        if value_lower is not None:
            if include_equal_to_lower is True:
                clauses.append(column >= value_lower)
            else:
                clauses.append(column > value_lower)
        if value_upper is not None:
            if include_equal_to_upper is True:
                clauses.append(column <= value_upper)
            else:
                clauses.append(column < value_upper)


def add_date_range_clause(clauses, column, date_lower, date_upper,
                          include_equal_to_lower=True,
                          include_equal_to_upper=True,
                          **options):
    """
    adds date range comparison into given clauses using specified inputs.

    :param list clauses: clause list to add range clause to it.
    :param Column column: entity column to add range clause for it.
    :param datetime date_lower: lower bound of range clause.
    :param datetime date_upper: upper bound of range clause.

    :param include_equal_to_lower: specifies that lower date
                                   should be considered in range.
                                   defaults to True if not provided.

    :param include_equal_to_upper: specifies that upper date
                                   should be considered in range.
                                   defaults to True if not provided.

    :keyword bool consider_begin_of_day: specifies that consider begin
                                         of day for lower date.
                                         defaults to True if not provided.

    :keyword bool consider_end_of_day: specifies that consider end
                                       of day for upper date.
                                       defaults to True if not provided.
    """

    consider_begin_of_day = options.get('consider_begin_of_day', True)
    consider_end_of_day = options.get('consider_end_of_day', True)

    if date_lower is not None and consider_begin_of_day is True:
        date_lower = datetime_utils.begin_of_day(date_lower)

    if date_upper is not None and consider_end_of_day is True:
        date_upper = datetime_utils.end_of_day(date_upper)

    add_range_clause(clauses, column, date_lower, date_upper,
                     include_equal_to_lower, include_equal_to_upper)


def add_list_clause(clauses, column, value):
    """
    adds list or single comparison into clauses based on given value.

    :param list clauses: clause list to add comparison clause to it.
    :param Column column: entity column to add comparison clause for it.
    :param Union[object, list[object]] value: value to add comparison for it.
    """

    if value is not None:
        if isinstance(value, LIST_TYPES):
            clauses.append(column.in_(value))
        else:
            clauses.append(column == value)


def add_like_clause(clauses, column, value, string_wrapper=like_both):
    """
    adds like clause into clauses based on given inputs.

    :param list clauses: clause list to add like clause to it.
    :param Column column: entity column to add like clause for it.
    :param object value: value to add like clause for it.
    :param callable string_wrapper: a callable to provide a string
                                    to be used as value in like clause.
                                    defaults to `like_both` if not provided.
    """

    if value is not None:
        clauses.append(column.like(string_wrapper(value)))


def create_row_result(fields, values):
    """
    creates a row result object with given fields and values.
    this object type is returned by sqlalchemy `Query` when there
    is columns or multiple entities in query.

    :param list[str] fields: field names of the result object.

    :param list[object] values: values to be mapped to fields.
                                they must be in the same order as fields.

    :raises InvalidRowResultFieldsAndValuesError: invalid row result fields
                                                  and values error.

    :raises FieldsAndValuesCountMismatchError: fields and values count mismatch error.

    :rtype: AbstractKeyedTuple
    """

    if fields is None or values is None:
        raise InvalidRowResultFieldsAndValuesError('Input parameters "fields" and '
                                                   '"values" must both be provided, '
                                                   'they could not be None.')

    if len(fields) != len(values):
        raise FieldsAndValuesCountMismatchError('The length of "fields" which is '
                                                '[{fields}] and "values" which is '
                                                '[{values}] does not match.'
                                                .format(fields=len(fields),
                                                        values=len(values)))

    keyed_tuple = lightweight_named_tuple('result', fields)
    result = keyed_tuple(values)

    return result



# -*- coding: utf-8 -*-
"""
database orm sql services module.
"""

from pyrin.application.services import get_component
from pyrin.database.orm.sql import DatabaseORMSQLPackage


def find_table_names(expression, **options):
    """
    finds table names from a string or `TextClause` sql expression.

    :param Union[str, TextClause] expression: a string or `TextClause`
                                              containing a sql expression.

    :keyword bool include_select: specifies that select statements
                                  must be investigated for table names.
                                  defaults to True if not provided.

    :keyword bool include_insert: specifies that insert statements
                                  must be investigated for table names.
                                  defaults to True if not provided.

    :keyword bool include_update: specifies that update statements
                                  must be investigated for table names.
                                  defaults to True if not provided.

    :keyword bool include_delete: specifies that delete statements
                                  must be investigated for table names.
                                  defaults to True if not provided.

    :returns: list[str]
    :rtype: list
    """

    return get_component(DatabaseORMSQLPackage.COMPONENT_NAME).find_table_names(expression,
                                                                                **options)



# -*- coding: utf-8 -*-
"""
token package.
"""

from pyrin.packaging import Package


class TokenPackage(Package):
    """
    token package class.
    """

    NAME = __name__
    DEPENDS = []
    COMPONENT_NAME = 'security.token.component'
